
# Den her kode base er en simpel version af en game engine.
# Den er skrevet i Python og bruger Pyglet til at håndtere vinduer og grafik.

import pyglet as pgl
from typing import DefaultDict
import numpy as np
import math
from time import time

###########################################
# Engine
# TODO:
# - Smooth scene transitions
###########################################
class Engine:
    def __init__(self, width=600, height=600):
        self._width = width
        self._height = height
        
        self.window = pgl.window.Window(self._width, self._height)
        self._currentScene = None
        self._keyboard = pgl.window.key.KeyStateHandler()
        self.__keyNames = {v: k for k, v in pgl.window.key._key_names.items()}

        self._colliders = []
        self._physicsObjects = []
        self._flags = {"debug":False}

        self.last = time()
        self.frames = 0
        self.fps_text = pgl.text.Label(text='Unknown', font_name='Verdana', font_size=8, x=10, y=10, color=(255,255,255,255))


    # flags
    def setFlag(self, flag:str, value:bool) -> None:
        self._flags[flag] = value

    def getFlag(self, flag:str) -> bool:
        return self._flags[flag]


    # event handling
    def isKeyPressed(self, key:str) -> bool:

        keyval = self.__keyNames.get(key.upper())

        if keyval is None:
            return False

        return self._keyboard[keyval]
        
    
    # collider handling
    def addCollider(self, collider):
        self._colliders.append(collider)
    

    # add physics object
    def addPhysicsObject(self, physicsObject):
        self._physicsObjects.append(physicsObject)

    # get collisions O(1)
    def isColliding(self, collider1, collider2) -> bool:
        return self.collisions[collider1].__contains__(collider2)

    
    # scene handling
    def setCurrentScene(self, scene):
        self._currentScene = scene
        self._colliders = []
        self._physicsObjects = []
        scene.setup()


    def __calculateCollisions__(self):
        
        self.collisions = DefaultDict(set)

        rows = 100
        cols = 100

        w, h = self._width/rows + 1, self._height/cols + 1

        matrix = [[[] for _ in range(cols)] for _ in range(rows)]

        # divide
        for collider in self._colliders:
            
            x0 = int(collider.x / w)
            y0 = int(collider.y / h)

            x1 = int((collider.x + collider.width) / w) + 1
            y1 = int((collider.y + collider.height) / h) + 1

            for x in range(x0, x1):
                for y in range(y0, y1):
                    
                    if x < 0 or x >= cols or y < 0 or y >= rows:
                        continue

                    matrix[x][y].append(collider)

        # check collision
        for x in range(cols):
            
            for y in range(rows):

                colliders = matrix[x][y]

                for i in range(len(colliders)):
                    for j in range(i+1, len(colliders)):
                        if colliders[i].isColliding(colliders[j]):
                            self.collisions[colliders[i]].add(colliders[j])
                            self.collisions[colliders[j]].add(colliders[i])

    # loop
    def __update__(self, dt):
        
        self.__calculateCollisions__()

        for physicsObject in self._physicsObjects:
            physicsObject.update(self, dt)

        # i få situationer er der ikke sat en scene
        if self._currentScene:
            self._currentScene.update(dt)

    
    def __draw__(self):
        self.window.clear()
        
        if self._flags["debug"]:
        
            # draw fps
            if time() - self.last >= 1:
                self.fps_text.text = str(self.frames)
                self.frames = 0
                self.last = time()
            else:
                self.frames += 1
            self.fps_text.draw()

        # i få situationer er der ikke sat en scene
        if self._currentScene:
            self._currentScene.draw()


    def start(self):
        
        self.window.on_draw = self.__draw__

        self.window.push_handlers(self._keyboard)
        pgl.clock.schedule_interval(self.__update__, 1/128)
        pgl.app.run()







# Classical game engine
from .engine import Engine
from .player import ControllablePlayer
from .collider import Collider
from .physics import PhysicsObject
from .animation import Animations

# Individual learning games


# Other
# Gør det nemere at impotere for læring
# NOTE: Can of course create more namespace
from pyglet.shapes import Rectangle, Circle, Line, Arc, Ellipse, Triangle, Polygon, BezierCurve




from .collider import Collider
import numpy as np

###########################################
# PhysicsObject
###########################################
class PhysicsObject:
    """
    Physicsobjektets eneste mål er at ændre self.x og self.y på objektet

    Husk at dette objekt IKKE ændrer hvad der tegnes. Det skal man selv gøre

    """

    # 100 px svarer til 1 meter

    def __init__(self, object, collider:Collider, mass=100, immovable=False, drag=1):
        self.collider = collider # use objects collider
        self.collider._physicsObject = self # add reference to physics object in collider
        self.object = object
        self.immovable = immovable

        # add x and y to objects
        if not hasattr(object, 'x') or not hasattr(object, 'y'):
            print(object, 'har ingen x eller y. Dette tilføjes nu')
            object.x = collider.x
            object.y = collider.y

        self.collider._lastpos = [collider.x, collider.y]

        self.mass = mass

        self.velocity = np.array([0.0,0.0])
        self.forces = np.array([0.0,0.0])

        self.drag = drag * mass
        

    def addForce(self, x, y):
        self.forces += np.array([x, y])
    

    def update(self, engine, dt):

        # calculate drag
        if abs(self.velocity[0]) > 0:
            self.forces[0] -= self.drag * self.velocity[0]
        
        if abs(self.velocity[1]) > 0:
            self.forces[1] -= self.drag * self.velocity[1]
        
        # calculate velocity
        self.velocity += self.forces / self.mass * dt

        # reset forces
        self.forces = np.array([0.0, 0.0])

        # calculate inertia
        for collider in engine.collisions[self.collider]:
            
            if not hasattr(collider, '_physicsObject'):
                continue

            physics:PhysicsObject = collider._physicsObject

            velIndex, direction = self.__checkWhichSideCollision__(collider)

            if velIndex == None:
                continue

            # laver kun beregning for største objekt
            if physics.mass > self.mass:
                                
                va1 = self.velocity[velIndex]
                vb1 = physics.velocity[velIndex]

                va = ((self.mass - physics.mass) * va1 + (2 * physics.mass) * vb1) / (self.mass + physics.mass)
                vb = ((2 * self.mass) * va1 + (physics.mass - self.mass) * vb1) / (self.mass + physics.mass)

                self.velocity[velIndex] = va
                physics.velocity[velIndex] = vb

        # update position
        self.object.x += (self.velocity[0] * dt) * 1000
        self.object.y += (self.velocity[1] * dt) * 1000

        # force borders
        for collider in engine.collisions[self.collider]:
        
            # if physics.mass > self.mass:
                # fjern overlap
                if velIndex == 0:

                    if direction == 1:
                        self.object.x = min(self.object.x + self.collider.width, physics.collider.x) - self.collider.width
                    
                    if direction == -1:
                        self.object.x = max(self.object.x, physics.collider.x + physics.collider.width)

                else:
                        
                    if direction == 1:
                        self.object.y = min(self.object.y + self.collider.height, physics.collider.y) - self.collider.height
                    
                    if direction == -1:
                        self.object.y = max(self.object.y, physics.collider.y + physics.collider.height)

        # update last position
        if self.immovable:
            self.object.x = self.collider._lastpos[0]
            self.object.y = self.collider._lastpos[1]
            self.velocity = np.array([0.0,0.0])
        
        self.collider.x = self.object.x
        self.collider.y = self.object.y

        self.collider._lastpos = [self.object.x, self.object.y]


    def __checkWhichSideCollision__(self, other):
        """
        returns (velocity index, direction)
        """

        self.width = self.collider.width
        self.height = self.collider.height

        ox, oy = other._lastpos
        sx, sy = self.collider._lastpos

        dx = (sx + self.width / 2) - (ox + other.width / 2)
        dy = (sy + self.height / 2) - (oy + other.height / 2)
        width = (self.width + other.width) / 2
        height = (self.height + other.height) / 2
        crossWidth = width * dy
        crossHeight = height * dx

        if abs(dx) <= width and abs(dy) <= height:
            if crossWidth > crossHeight:
                if crossWidth > -crossHeight:
                    return (1, -1) #bottom
                else:
                    return (0, 1) #left
            else:
                if crossWidth > -crossHeight:
                    return (0, -1) # right
                else:
                    return (1, 1) # top
        return None, None




from .engine import Engine
from .collider import Collider
from .physics import PhysicsObject
from pyglet.shapes import Rectangle

class ControllablePlayer:

    def __init__(self, engine:Engine, x=300, y=300, w=50, h=50, mass=60):
        self.x = y
        self.y = x
        self.w = w
        self.h = h

        self.rect = Rectangle(x, y, w, h, color=(198, 39, 114))
        self.collider = Collider(x, y, w, h, self)
        self.physics = PhysicsObject(self, self.collider, mass)

        # reference
        self.engine = engine


    def __update__(self):
        self.rect.x = self.x
        self.rect.y = self.y


    def draw(self):
        self.rect.draw()


    def update(self):
        
        force = 100

        # bevæg den
        if self.engine.isKeyPressed('w'):
            self.physics.addForce(0, force)

        if self.engine.isKeyPressed('s'):
            self.physics.addForce(0, -force)
        
        if self.engine.isKeyPressed('a'):
            self.physics.addForce(-force, 0)
        
        if self.engine.isKeyPressed('d'):
            self.physics.addForce(force, 0)

        self.__update__()





import pyglet as pgl

###########################################
# Collider 
###########################################
class Collider:
    
    def __init__(self, x, y, width, height, referenceObject):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.rect = pgl.shapes.Rectangle(x, y, width, height, color=(255, 0, 0))

        self.referenceObject = referenceObject


    def draw(self):
        self.rect.x = self.x
        self.rect.y = self.y

        self.rect.draw()


    def isColliding(self, other):
        isColliding = not (self.x + self.width < other.x or
                self.x > other.x + other.width or
                self.y + self.height < other.y or
                self.y > other.y + other.height)

        if isColliding:
            self.rect.color = (255, 0, 0)
        else:
            self.rect.color = (0, 0, 255)

        return isColliding




"""

References

https://pyglet.readthedocs.io/en/latest/programming_guide/image.html#animations

For some reason binding animations to texture atlas creates problems

"""

import pyglet as pgl
from PIL import Image


class Animations:

    
    def __addAnimationState__(self, 
                              fullSpriteSheet, 
                              size, 
                              y0, 
                              y1, 
                              cols, 
                              period, 
                              rowHeight,
                              flipx=False,
                              flipy=False
                              ):

        # chop images into diffrent textures
        spriteSheet = fullSpriteSheet.crop([0, y0, fullSpriteSheet.width, y1])
        rawImage = spriteSheet.tobytes()  # tostring is deprecated
        spriteSheet = pgl.image.ImageData(spriteSheet.width, spriteSheet.height, 'RGBA', rawImage, pitch=-spriteSheet.width * 4)

        # load image into image grid and add animation
        imageGrid = pgl.image.ImageGrid(spriteSheet, rows=1, columns=cols)
        
        animation = imageGrid.get_animation(period=period)
        
        for frame in animation.frames:
            texture = frame.image.get_texture()
            frame.image = texture.get_transform(flipx, flipy)

        sprite = pgl.sprite.Sprite(img=animation)

        sprite.scale = size/rowHeight

        return sprite


    def __init__(self, texturePath:str, rows:int, cols:int, period=1/15, size=100):
        """
        uses center pos
        """

        self.size = size


        fullSpriteSheet = Image.open(texturePath)
        rowHeight = fullSpriteSheet.height / rows
        
        self.cols = cols
        self.rows = rows

        self.animations = {
            "default":[],
            "flipx":[],
            "flipy":[]
        }

        self._forceNext = None
        self._forceNextActive = False

        for i in range(rows):

            self.animations["default"].append(self.__addAnimationState__(
                fullSpriteSheet, 
                size,
                rowHeight*i, 
                rowHeight*(i+1), 
                cols, 
                period, 
                rowHeight
            ))

        for i in range(rows):

            self.animations["flipx"].append(self.__addAnimationState__(
                fullSpriteSheet, 
                size,
                rowHeight*i, 
                rowHeight*(i+1), 
                cols, 
                period, 
                rowHeight,
                flipx=True
            ))
        
        for i in range(rows):

            self.animations["flipy"].append(self.__addAnimationState__(
                fullSpriteSheet, 
                size,
                rowHeight*i, 
                rowHeight*(i+1), 
                cols, 
                period, 
                rowHeight,
                flipy=True
            ))

        self._currentAnimation = [0, 'default']
        self._nextAnimationIndex = 0
        self._reachedEndOfCycle = False
        
        self.x = 0 # center pos
        self.y = 0


    def switch(self, index:int, flipx:bool=False, flipy:bool=False):
        """
        Bruges til states
        """
        
        if flipx and flipy:
            raise NotImplemented('Both flips can not happened at once')

        state = 'default'
        if flipx: state = 'flipx'
        if flipy: state = 'flipy'

        self._nextAnimationIndex = index
        self._currentAnimation[1] = state


    def do(self, index:int, flipx:bool=False, flipy:bool=False):
        
        self.switch(index, flipx, flipy)

        self._forceNext = self._nextAnimationIndex


    def setPos(self, x, y):
        self.x = x
        self.y = y


    def draw(self, x=None, y=None):

        if x is not None:
            self.x = x
        
        if y is not None:
            self.y = y


        index, state = self._currentAnimation

        sprite:pgl.sprite.Sprite = self.animations[state][index]
        sprite.x = x + (state == "flipx") * self.size - self.size/2
        sprite.y = y + (state == "flipy") * self.size - self.size/2

        # denne burde kun køre en gang per cyklus men bliver kørt flere gange 
        # hver gang der tegnes og man er på det sidste frame
        if sprite.frame_index == self.cols - 1 and not self._reachedEndOfCycle:
            
            # hvis der er en ny forced animation som ikke er igang
            if self._forceNext and not self._forceNextActive:
                # så skal den sættes på nu, da animations cyklussen er klar 
                self._forceNextActive = True
                self._currentAnimation[0] = self._forceNext
            
            # hvis den forced animation er der og har været igang
            elif self._forceNext and self._forceNextActive:
                # så skal den sættes fjerens nu da animationen er done
                self._forceNext = None
                self._forceNextActive = False
                self._currentAnimation[0] = self._nextAnimationIndex

            # hvis der ingen forced er
            else:
                self._currentAnimation[0] = self._nextAnimationIndex
            
        self._reachedEndOfCycle = sprite.frame_index == self.cols - 1

        sprite.draw()